// Solutions to the Advent Of Code 2024.
// Copyright (C) 2025  Stefan Müller
//
// This program is free software: you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.

#include <aoc/ChronospatialComputer.hpp>

#include <sstream>
#include <string>

ChronospatialComputer::ChronospatialComputer()
    : state_{}, program_{}, instructions_{
          std::make_unique<ChronospatialComputerDivisionInstruction>(0),   // adv
          std::make_unique<ChronospatialComputerXorLiteralInstruction>(),  // bxl
          std::make_unique<ChronospatialComputerModuloInstruction>(),      // bst
          std::make_unique<ChronospatialComputerJumpInstruction>(),        // jnz
          std::make_unique<ChronospatialComputerXorRegisterInstruction>(), // bxc
          std::make_unique<ChronospatialComputerOutInstruction>(),         // out
          std::make_unique<ChronospatialComputerDivisionInstruction>(1),   // bdv
          std::make_unique<ChronospatialComputerDivisionInstruction>(2)    // cdv
      }
{
}

const std::string ChronospatialComputer::getPuzzleName() const
{
    return "Chronospatial Computer";
}

const int ChronospatialComputer::getPuzzleDay() const
{
    return 17;
}

void ChronospatialComputer::processDataLine(const std::string& line)
{
    std::istringstream stream{ line };
    std::string token{};
    char c;
    int value;
    if (stream >> token)
    {
        if (token == "Register")
        {
            stream >> c >> token >> value;
            // c must be 'A', 'B', or 'C'.
            size_t regIndex{ static_cast<size_t>(c - 65) };
            state_.registers[regIndex] = value;
        }
        else
        {
            while (stream >> value)
            {
                program_.push_back(value);
                // Streams a comma from between values.
                stream >> c;
            }
        }
    }
}

void ChronospatialComputer::finish()
{
    runProgram(program_, state_);
    part1 = state_.output.str();
}

void ChronospatialComputer::runProgram(const std::vector<int>& program, ChronospatialComputerState& state) const
{
    state.instructionPointer = 0;
    while (state.instructionPointer < program.size())
    {
        instructions_[program[state.instructionPointer]]->run(state, program[state.instructionPointer + 1]);
    }
}
